#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <pwd.h>
#include <utmp.h>
#include <paths.h>
#include <syslog.h>

extern char **environ;

#define MAX_BUF 1024
char buffer[MAX_BUF];

/* Get virtual users password entry */
struct passwd *vs_getpwnam(const char *user, const char *passwdfile);

/*
	This program uses the GECOS field from real server accounts to determine
	which virtual server account to log into.

	UTMP/WTMP entries are updated.
*/
int main(int argc, char **argv) {
struct passwd *pw = getpwnam(getenv("LOGNAME"));
char gecos[sizeof(pw->pw_gecos) + 16], *vuser, *vserver;
char *args[argc + 16];
char *newargs[] = { "/usr/sbin/vserver", "--silent", NULL, "exec", "/bin/su", "-", NULL };
int i;
pid_t child;
char *tty = ttyname(0) ? ttyname(0) : ttyname(1);
struct stat tty_stat;

	openlog("vslogin", LOG_PID, LOG_AUTHPRIV);

	if (geteuid() != 0) {
		syslog(LOG_ERR, "vslogin is not setuid root");
		return 1;
	}

	strcpy(gecos, pw->pw_gecos);

	/* Stat TTY for later retrieval */
	if (tty)
		if (stat(tty, &tty_stat)) {
			syslog(LOG_ERR, "failed to stat TTY %s", tty);
			return 1;
		}

	for (i = 0; i < sizeof(newargs) / sizeof(*newargs); ++i)
		args[i] = newargs[i];

	vuser = gecos;
	if (!(vserver = strchr(gecos, '@'))) {
		syslog(LOG_ERR, "no vserver specified in GECOS *(%s)\n",gecos);
		return 1;
	}
	*vserver = 0;
	++vserver;

	/* Substitute in server and user */
	args[2] = vserver;
	args[6] = vuser;

	/* Copy existing args */
	for (i = 0; i < argc - 1; ++i)
		args[7 + i] = argv[i + 1];
	args[7 + i] = NULL;

	/*	Parent chroots into vserver, updates utmp/wtmp and waits for child to
		finish before cleaning up utmp/wtmp entries. */
	if ((child = fork())) {
	struct utmp entry, *real;

		memset(&entry, 0, sizeof(entry));

		/* Add wtmp entry */
		if (tty)
			strcpy(entry.ut_line, tty + strlen("/dev/"));

		/* Need to sleep to ensure previous UTMP entry has been written */
		sleep(1);

		/* Get UTMP entry from real server */
		setutent();
		if (tty && (real = getutline(&entry))) {
		struct passwd *pw = getpwuid(getuid());

			memset(&entry, 0, sizeof(entry));
			/*	TODO -- Remember that the vserver script must *exec* the
				subcommand, not merely call it. */
			entry.ut_pid = child;
			entry.ut_type = USER_PROCESS;
			strcpy(entry.ut_line, real->ut_line);
			strcpy(entry.ut_id, real->ut_id);
			if (!pw) {
				syslog(LOG_WARNING, "could not find logged in user?");
				strcpy(entry.ut_user, real->ut_user);
			} else {
			char vse[strlen(pw->pw_gecos) + 16];
			char *vserver;
			
				strcpy(vse, pw->pw_gecos);

				if (!(vserver = strchr(vse, '@')))
					syslog(LOG_WARNING, "user's GECOS field is incorrect (%s)", pw->pw_gecos);
				else {
					*vserver = 0;
					strcpy(entry.ut_user, vse);
				}
			}
			strcpy(entry.ut_host, real->ut_host);
			time(&entry.ut_time);
			endutent();

			/* Enter vserver root */
			snprintf(buffer, MAX_BUF, "/vservers/%s", vserver);
			if (chroot(buffer)) perror("vslogin:chroot");

			/* Add UTMP entry */
			setutent();
			if (!pututline(&entry)) {
				syslog(LOG_ERR, "could not add UTMP entry to vserver; %s", strerror(errno));
				return 1;
			}
			endutent();

			updwtmp(_PATH_WTMP, &entry);

			/* Wait for child */
			waitpid(child, NULL, 0);

			/* Clear UTMP entry */
			memset(&entry, 0, sizeof(entry));
			entry.ut_type = DEAD_PROCESS;
			if (tty)
				strcpy(entry.ut_line, tty + strlen("/dev/"));
			time(&entry.ut_time);

			if (!pututline(&entry)) {
				syslog(LOG_ERR, "could not clear UTMP entry from vserver; %s", strerror(errno));
				return 1;
			}
			endutent();
			updwtmp(_PATH_WTMP, &entry);

			/* Chown TTY back to original owner and permissions */
			if (tty) {
				chown(tty, tty_stat.st_uid, tty_stat.st_gid);
				chmod(tty, tty_stat.st_mode);
			}
		} else {
			syslog(LOG_WARNING, "failed to find parent tty");
			waitpid(child, NULL, 0);
		}
		return 0;
	}

	/* Change ownership of TTY to that of virtual user */
	if (tty) {
		if (stat(tty, &tty_stat)) {
			syslog(LOG_ERR, "failed to stat TTY %s", tty);
			return 1;
		}

		snprintf(buffer, MAX_BUF, "/vservers/%s/etc/passwd", vserver);
		if (!(pw = vs_getpwnam(vuser, buffer))) {
			syslog(LOG_ERR, "failed to get passwd entry %s from vserver %s", vuser, vserver);
			return 1;
		}

		if (chown(tty, pw->pw_uid, tty_stat.st_gid)) {
			syslog(LOG_ERR, "failed to chown TTY %s to %d", tty, pw->pw_uid);
			return 1;
		}
	}

	setuid(geteuid());
	if (execv("/usr/sbin/vserver", args)) {
		syslog(LOG_ERR, "failed to execv /usr/sbin/vserver; %s", strerror(errno));
		return 2;
	}
	return 0;
}

struct passwd *vs_getpwnam(const char *user, const char *passwd) {
FILE *f = fopen(passwd, "r");
struct passwd *pw;

	while ((pw = fgetpwent(f))) {
		if (!strcmp(pw->pw_name, user))
			return pw;
	}
	return NULL;
}
